"""
This file contains glossary-related object_action functions for processing Egeria Markdown
"""

import json
import os
import re
from typing import List, Optional

from rich import print
from rich.console import Console
from rich.markdown import Markdown

from md_processing.md_processing_utils.common_md_utils import (debug_level, print_msg, set_debug_level,
                                                               get_element_dictionary, update_element_dictionary)
from md_processing.md_processing_utils.extraction_utils import (extract_command_plus, extract_command,
                                                                process_simple_attribute, process_element_identifiers,
                                                                update_a_command, extract_attribute,
                                                                get_element_by_name, process_name_list)
from md_processing.md_processing_utils.md_processing_constants import (GLOSSARY_NAME_LABELS, TERM_NAME_LABELS,
                                                                       TERM_RELATIONSHPS, PARENT_CATEGORY_LABELS,
                                                                       CATEGORY_NAME_LABELS, ALWAYS, ERROR, INFO,
                                                                       WARNING, pre_command, EXISTS_REQUIRED,
                                                                       OUTPUT_LABELS, SEARCH_LABELS, GUID_LABELS,
                                                                       ELEMENT_OUTPUT_FORMATS, command_seperator)
from pyegeria import body_slimmer
from pyegeria._globals import (NO_GLOSSARIES_FOUND, NO_ELEMENTS_FOUND, NO_CATEGORIES_FOUND)
from pyegeria.egeria_tech_client import EgeriaTech

EGERIA_WIDTH = int(os.environ.get("EGERIA_WIDTH", "170"))
console = Console(width=EGERIA_WIDTH)


def update_term_categories(egeria_client: EgeriaTech, term_guid: str, categories_exist: bool,
                           categories_list: List[str]) -> None:
    """

    Adds or removes a term to/from specified categories in a glossary.

    This function associates a term, identified by its GUID, with one or more
    categories. It uses the provided EgeriaTech client to assign the term
    to the categories. If the GUID of a category is not readily available, it
    is retrieved either from a pre-loaded dictionary or through a client lookup.

    Args:
        egeria_client (EgeriaTech): The client to interact with the glossary.
        term_guid (str): The GUID of the term to be associated with categories.
        categories_exist (bool): Flag indicating whether the categories already
                                 exist.
        categories_list (List[str]): A list of category names to associate with
                                     the term.

    Returns:
        None
    """
    to_be_cat_guids: list[str] = []
    # find the categories a term is currently in.
    existing_categories = egeria_client.get_categories_for_term(term_guid)
    if type(existing_categories) is str:
        current_categories = []
    else:
        current_categories = [cat['elementHeader']['guid'] for cat in existing_categories]

    if categories_exist is True and categories_list is not None:
        if type(categories_list) is str:
            categories_list = categories_list.split(",").trim()
        for category in categories_list:
            cat_guid = None
            cat_el = category.strip()
            element_dict = get_element_dictionary()
            if cat_el in element_dict:
                cat = element_dict.get(cat_el, None)
                cat_guid = cat.get('guid', None) if cat else None
            if cat_guid is None:
                cat_guid = egeria_client.__get_guid__(qualified_name=cat_el)
                update_element_dictionary(cat_el, {'guid': cat_guid})
            to_be_cat_guids.append(cat_guid)

        for cat in to_be_cat_guids:
            if cat not in current_categories:
                egeria_client.add_term_to_category(term_guid, cat)
                current_categories.append(cat)
                msg = f"Added term {term_guid} to category {cat}"
                print_msg("DEBUG-INFO", msg, debug_level)

        for cat in current_categories:
            if cat not in to_be_cat_guids:
                egeria_client.remove_term_from_category(term_guid, cat)
                msg = f"Removed term {term_guid} from category {cat}"
                print_msg("DEBUG-INFO", msg, debug_level)
    else:  # No categories specified - so remove any categories a term is in
        for cat in current_categories:
            egeria_client.remove_term_from_category(term_guid, cat)
            msg = f"Removed term {term_guid} from category {cat}"
            print_msg("DEBUG-INFO", msg, debug_level)


# def update_term_categories(egeria_client: EgeriaTech, term_guid: str, current_categories: List[str],
# new_categories: List[str]) -> None:
#     """
#     Updates the categories of a term.
#
#     Args:
#         egeria_client: The Egeria client to use for the update.
#         term_guid: The GUID of the term to update.
#         current_categories: The current categories of the term.
#         new_categories: The new categories of the term.
#     """
#     if new_categories:  # If categories are specified, add them
#         for cat in new_categories:
#             if cat not in current_categories:
#                 egeria_client.add_term_to_category(term_guid, cat)
#                 msg = f"Added term {term_guid} to category {cat}"
#                 print_msg("DEBUG-INFO", msg, debug_level)
#         # Remove any categories that are not in the new list
#         for cat in current_categories:
#             if cat not in new_categories:
#                 egeria_client.remove_term_from_category(term_guid, cat)
#                 msg = f"Removed term {term_guid} from category {cat}"
#                 print_msg("DEBUG-INFO", msg, debug_level)
#     else:  # No categories specified - so remove any categories a term is in
#         for cat in current_categories:
#             egeria_client.remove_term_from_category(term_guid, cat)
#             msg = f"Removed term {term_guid} from category {cat}"
#             print_msg("DEBUG-INFO", msg, debug_level)


def process_glossary_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a glossary create or update object_action by extracting key attributes such as
    glossary name, language, description, and usage from the given text.

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """

    command, object_type, object_action = extract_command_plus(txt)
    set_debug_level(directive)

    glossary_name = process_simple_attribute(txt, GLOSSARY_NAME_LABELS, ERROR)
    print(Markdown(
        f"{pre_command} `{object_action}` `{object_type}`  for glossary: `\'{glossary_name}\'` with directive: `"
        f"{directive}` "))
    language = process_simple_attribute(txt, ['Language'], INFO)
    description = process_simple_attribute(txt, ['Description'], INFO)
    usage = process_simple_attribute(txt, ['Usage'], INFO)
    q_name = process_simple_attribute(txt, ['Qualified Name'], INFO)
    valid = True

    if glossary_name is None:
        valid = False
        known_q_name = None
        known_guid = None
        glossary_exists = False
    else:
        element_labels = GLOSSARY_NAME_LABELS
        element_labels.append('Display Name')
        known_q_name, known_guid, valid, glossary_exists = process_element_identifiers(egeria_client, object_type,
                                                                                       element_labels, txt,
                                                                                       object_action, None)
    glossary_display = (f"\n* Command: `{command}`\n\t* Glossary Name: {glossary_name}\n\t"
                        f"* Language: {language}\n\t* Description:\n{description}\n"
                        f"* Usage: {usage}\n")

    if object_action == 'Update':
        guid = process_simple_attribute(txt, ['GUID', 'guid', 'Guid'])
        glossary_display += f"* Qualified Name: `{q_name}`\n\t* GUID: {guid}\n\n"
        if not glossary_exists:
            msg = f"Glossary can't be updated; `{glossary_name}` not found"
            print_msg("ERROR", msg, debug_level)
            valid = False
        else:
            msg = f"Glossary can be updated; `{glossary_name}` found"
            print_msg(ALWAYS, msg, debug_level)


    elif object_action == "Create":
        if glossary_exists:
            msg = f"Glossary `{glossary_name}` can't be created because it already exists.\n"
            print_msg("ERROR", msg, debug_level)
            valid = False
        elif valid:
            msg = f"It is valid to create Glossary `{glossary_name}` with:\n"
            print_msg("ALWAYS", msg, debug_level)

    if directive == "display":
        print(Markdown(glossary_display))
        return None

    elif directive == "validate":
        if valid:
            print(Markdown(glossary_display))
        else:
            msg = f"Validation failed for Glossary `{glossary_name}`\n"
            print_msg(ERROR, msg, debug_level)
            print(Markdown(glossary_display))

        return valid

    elif directive == "process":
        if valid:
            print(Markdown(glossary_display))
        else:
            if glossary_exists and object_action == "Create":
                msg = (f"Create failed because glossary `{glossary_name}` exists - changing `Create` to `Update` in "
                       f"processed output \n")
                print_msg(ERROR, msg, debug_level)
                print(Markdown(glossary_display))
                print(command_seperator)
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)
            else:
                return None
        if object_action == "Update":
            if not glossary_exists:
                print(f"\n{ERROR}Glossary `{glossary_name}` does not exist! Updating result document with Create "
                      f"object_action\n")
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)

            body = {
                "class": "ReferenceableRequestBody", "elementProperties": {
                    "class": "GlossaryProperties", "qualifiedName": known_q_name, "description": description,
                    "language": language, "usage": usage
                    }
                }
            egeria_client.update_glossary(known_guid, body)
            print_msg(ALWAYS, f"Updated Glossary `{glossary_name}` with GUID {known_guid}", debug_level)
            update_element_dictionary(known_q_name, {
                'guid': known_guid, 'display_name': glossary_name
                })
            return egeria_client.get_glossary_by_guid(known_guid, output_format='MD')
        elif object_action == "Create":
            glossary_guid = None

            if glossary_exists:
                print(f"\nGlossary `{glossary_name}` already exists and result document updated\n")
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)
            else:
                glossary_guid = egeria_client.create_glossary(glossary_name, description, language, usage)
                glossary = egeria_client.get_glossary_by_guid(glossary_guid)
                if glossary == NO_GLOSSARIES_FOUND:
                    print(f"{ERROR}Just created with GUID {glossary_guid} but Glossary not found\n")
                    return None
                qualified_name = glossary['glossaryProperties']["qualifiedName"]
                update_element_dictionary(qualified_name, {
                    'guid': glossary_guid, 'display_name': glossary_name
                    })
                # return update_a_command(txt, object_action, object_type, qualified_name, glossary_guid)
                print_msg(ALWAYS, f"Created Glossary `{glossary_name}` with GUID {glossary_guid}", debug_level)
                return egeria_client.get_glossary_by_guid(glossary_guid, output_format='FORM')
        else:
            return None
    else:
        return None


def process_category_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a glossary category create or update object_action by extracting key attributes such as
    category name, qualified, description, and anchor glossary from the given txt..

    :param txt: A string representing the input cell to be processed for
        extracting category-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    valid = True
    set_debug_level(directive)

    command, object_type, object_action = extract_command_plus(txt)

    category_name = process_simple_attribute(txt, ['Category Name', 'category_name', 'Cat'])
    print(Markdown(f"{pre_command} `{command}` for category: `\'{category_name}\'` with directive: `{directive}` "))

    owning_glossary_name = extract_attribute(txt, ['Owning Glossary', 'In Glossary'])
    description = process_simple_attribute(txt, ['Description'])
    q_name = process_simple_attribute(txt, ['Qualified Name'])

    parent_category_name = process_simple_attribute(txt, PARENT_CATEGORY_LABELS, "INFO")

    element_labels = CATEGORY_NAME_LABELS
    element_labels.append('Display Name')
    # Check if category exists (and get qname and guid)
    if category_name is None:
        valid = False
        known_q_name, known_guid, category_exists = None
    else:
        element_labels = CATEGORY_NAME_LABELS
        element_labels.append('Display Name')
        known_q_name, known_guid, valid, category_exists = process_element_identifiers(egeria_client, object_type,
                                                                                       element_labels, txt,
                                                                                       object_action, None)

    # Check if owning glossary exists (and get qname)
    if owning_glossary_name is None:
        valid = False
        known_glossary_q_name, known_glossary__guid, glossary_exists = None

    else:
        known_glossary_q_name, known_glossary_guid, valid, owning_glossary_exists = process_element_identifiers(
            egeria_client, "Glossary", GLOSSARY_NAME_LABELS, txt, EXISTS_REQUIRED, None)

    if parent_category_name:
        _, parent_guid, parent_valid, parent_exists = get_element_by_name(egeria_client, 'Glossary Categories',
                                                                          parent_category_name)
    else:
        parent_guid = None
        parent_exists = False
        parent_valid = False

    category_display = (
        f"\n* Command: {command}\n\t* Category: {category_name}\n\t* In Glossary: {owning_glossary_name}\n\t"
        f"* Description:\n{description}\n\t* Parent Category: {parent_category_name}\n\t"
        f"* Qualified Name: {q_name}\n\t")

    if object_action == 'Update':
        guid = process_simple_attribute(txt, ['GUID', 'guid', 'Guid'])

        category_display += (f"* GUID: {guid}\n\n")
        if not category_exists:
            msg = f"Category {category_name} can't be updated; {category_name} not found."
            print_msg(ERROR, msg, debug_level)
            valid = False
        else:
            msg = f"Glossary can be updated; {category_name} found"
            print_msg(ALWAYS, msg, debug_level)

    elif object_action == "Create":
        if category_exists:
            msg = f"Category {category_name} can't be created because it already exists.\n"
            print_msg("ERROR", msg, debug_level)
            valid = False
        elif valid:
            msg = f"It is valid to create Category `{category_name}` with:\n"
            print_msg("ALWAYS", msg, debug_level)

    if directive == "display":
        print(Markdown(category_display))
        return None

    elif directive == "validate":
        if valid:
            print(Markdown(category_display))
        else:
            msg = f"Validation failed for {object_type} `{category_name}`\n"
            print_msg(ERROR, msg, debug_level)
            print(Markdown(category_display))
        return valid

    elif directive == "process":
        if valid:
            print(Markdown(category_display))
        else:
            if category_exists and object_action == "Create":
                msg = (f"Create failed because category `{category_name}` exists - changing `Create` to `Update` in "
                       f"processed output \n")
                print_msg(ERROR, msg, debug_level)
                print(Markdown(category_display))
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)
            else:
                return None

        if object_action == "Update":
            if not category_exists:
                print(f"\n{ERROR}category `{category_name}` does not exist! Updating result document with Create "
                      f"object_action\n")
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)

            # Update the basic category properties
            egeria_client.update_category(known_guid, category_name, description, known_q_name, None)
            msg = f"->Updated category `{category_name}`with GUID {known_guid}"
            print_msg(ALWAYS, msg, debug_level)

            # Update parent-child relationships

            update_element_dictionary(known_q_name, {
                'guid': known_guid, 'display_name': category_name
                })

            category_sync = update_category_parent(egeria_client, known_guid, parent_category_name)
            print_msg(ALWAYS, f"Updated Category hierarchy for  `{category_name}` with outcome {category_sync}",
                      debug_level)
            return egeria_client.get_category_by_guid(known_guid, output_format='FORM')

        elif object_action == "Create":
            is_root = True

            if category_exists:
                msg = (f"Cannot create`{category_name}` because it already exists; result document written for "
                       f"category update\n")
                print_msg(WARNING, msg, debug_level)
                return update_a_command(txt, object_action, object_type, known_q_name, known_guid)
            else:
                category_guid = egeria_client.create_category(known_glossary_guid, category_name, description, is_root)
                category_details = egeria_client.get_category_by_guid(category_guid)

                if category_details == NO_CATEGORIES_FOUND:
                    msg = f"Just created category with GUID {category_guid} but category not found"
                    print_msg(ERROR, msg, debug_level)
                    return None

                qualified_name = category_details['glossaryCategoryProperties']["qualifiedName"]
                update_element_dictionary(qualified_name, {
                    'guid': category_guid, 'display_name': category_name
                    })
                print_msg(ALWAYS, f"Created Category `{category_name}` with GUID {category_guid}", debug_level)
                if parent_valid and parent_guid:
                    egeria_client.set_parent_category(parent_guid, category_guid)
                    print_msg(ALWAYS, f"Set parent category for `{category_name}` to `{parent_category_name}`",
                              debug_level)
                else:
                    print_msg(ERROR,
                              f"Parent category `{parent_category_name}` not found or invalid for `{category_name}`",
                              debug_level)
                return egeria_client.get_category_by_guid(category_guid, output_format='FORM')
        return None
    return None


def update_category_parent(egeria_client, category_guid: str, parent_category_name: str = None) -> bool:
    """
    Updates the parent relationship for a category.

    If a parent category is specified, it will check if a parent is currently set.
    If a parent category was set and is the same as the parent category specified, no change is needed.
    If a parent category was set and is different from the parent_category_name, the parent category is updated.
    If parent_category_name is None or empty and an existing parent category was set, the parent category is removed.

    Parameters
    ----------
    egeria_client: EgeriaTech
        The Egeria client to use for API calls
    category_guid: str
        The GUID of the category to update
    parent_category_name: str, optional
        The name of the parent category to set, or None to remove the parent

    Returns
    -------

    True if successful, False otherwise.

    """
    outcome = True
    # Handle parent category updates
    if parent_category_name:
        # Check if a parent is currently set
        current_parent = egeria_client.get_category_parent(category_guid)

        if isinstance(current_parent, str) and "No Parent Category found" in current_parent:
            # No parent currently set, need to set it
            _, parent_guid, _, parent_exists = get_element_by_name(egeria_client, 'Glossary Categories',
                                                                   parent_category_name)

            if parent_exists and parent_guid:
                egeria_client.set_parent_category(parent_guid, category_guid)
                print_msg(ALWAYS, f"Set parent category of category to `{parent_category_name}`", debug_level)
            else:
                print_msg(ERROR, f"Parent category `{parent_category_name}` not found", debug_level)
                outcome = False
        else:
            # Parent is set, check if it's the same
            current_parent_name = current_parent.get('glossaryCategoryProperties', {}).get('qualifiedName', '')

            if current_parent_name != parent_category_name:
                # Different parent, need to update
                # First remove the current parent
                current_parent_guid = current_parent.get('elementHeader', {}).get('guid', '')
                if current_parent_guid:
                    egeria_client.remove_parent_category(current_parent_guid, category_guid)

                # Then set the new parent
                _, parent_guid, _, parent_exists = get_element_by_name(egeria_client, 'Glossary Categories',
                                                                       parent_category_name)

                if parent_exists and parent_guid:
                    egeria_client.set_parent_category(parent_guid, category_guid)
                    print_msg(ALWAYS,
                              f"Updated parent category from `{current_parent_name}` to `{parent_category_name}`",
                              debug_level)
                else:
                    print_msg(ERROR, f"Parent category `{parent_category_name}` not found", debug_level)
                    outcome = False
    elif parent_category_name is None or parent_category_name == '':
        # Check if a parent is currently set and remove it if needed
        current_parent = egeria_client.get_category_parent(category_guid)

        if not isinstance(current_parent, str) or "No Parent Category found" not in current_parent:
            # Parent is set, need to remove it
            current_parent_guid = current_parent.get('elementHeader', {}).get('guid', '')
            current_parent_name = current_parent.get('glossaryCategoryProperties', {}).get('qualifiedName', '')

            if current_parent_guid:
                egeria_client.remove_parent_category(current_parent_guid, category_guid)
                print_msg(ALWAYS, f"Removed parent category `{current_parent_name}`", debug_level)

    return outcome


def process_term_upsert_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """
    Processes a term create or update object_action by extracting key attributes such as
    term name, summary, description, abbreviation, examples, usage, version, and status from the given cell.

    :param txt: A string representing the input cell to be processed for
        extracting glossary-related attributes.
    :param directive: an optional string indicating the directive to be used - display, validate or execute
    :return: A string summarizing the outcome of the processing.
    """
    valid = True
    categories_list = None
    cats_exist = False
    set_debug_level(directive)
    known_q_name = None
    command = extract_command(txt)
    object_type = command.split(' ')[1].strip()
    object_action = command.split(' ')[0].strip()

    term_name = process_simple_attribute(txt, ['Term Name', 'Display Name'], ERROR)
    print(Markdown(f"{pre_command} `{command}` for term:`{term_name}` with directive: `{directive}`"))
    summary = process_simple_attribute(txt, ['Summary'], INFO)
    description = process_simple_attribute(txt, ['Description'], INFO)
    abbreviation = process_simple_attribute(txt, ['Abbreviation'], INFO)
    examples = process_simple_attribute(txt, ['Examples'], INFO)
    usage = process_simple_attribute(txt, ['Usage'], INFO)
    status = process_simple_attribute(txt, ['Status'])
    status = status.upper() if status else 'DRAFT'
    version = process_simple_attribute(txt, ['Version', "Version Identifier", "Published Version"], INFO)
    q_name = process_simple_attribute(txt, ['Qualified Name'], INFO)

    aliases = process_simple_attribute(txt, ['Aliases', 'Alias'], INFO)
    if aliases:
        alias_list = list(filter(None, re.split(r'[,\n]+', aliases.strip())))
    else:
        alias_list = None

    # validate term name and get existing qualified_name and guid if they exist
    if term_name is None:
        valid = False
        known_q_name, known_guid, term_exists = None
    else:
        element_labels = TERM_NAME_LABELS
        element_labels.append('Display Name')
        known_q_name, known_guid, valid, term_exists = process_element_identifiers(egeria_client, object_type,
                                                                                   element_labels, txt, object_action,
                                                                                   version)

    # get the glossary qualified name this term is in
    glossary_name = process_simple_attribute(txt, GLOSSARY_NAME_LABELS, ERROR)
    if glossary_name is None:
        valid = False
        known_glossary_guid = None
        known_glossary_q_name = None
        glossary_valid = False
        glossary_exists = False
    else:
        known_glossary_q_name, known_glossary_guid, glossary_valid, glossary_exists = process_element_identifiers(
            egeria_client, "Glossary", GLOSSARY_NAME_LABELS, txt, EXISTS_REQUIRED, None)

    # process categories, if present
    categories = process_simple_attribute(txt, ['Glossary Categories', 'Glossary Category', 'Category', 'Categories'])
    if categories:  # Find information about categoriess that classify this term
        msg = "Checking for categories that classify this term"
        print_msg("DEBUG-INFO", msg, debug_level)
        categories_list, cat_q_name_list, cats_valid, cats_exist = process_name_list(egeria_client,
                                                                                     'Glossary Categories', txt,
                                                                                     CATEGORY_NAME_LABELS)
        if cats_exist and cats_valid:
            msg = f"Found valid glossary categories to classify the term:\n\t{term_name}"
            print_msg("INFO", msg, debug_level)
        else:
            msg = "No valid glossary categories found."
            print_msg("INFO", msg, debug_level)
    else:
        cats_exist = cats_valid = False
        cat_q_name_list = None

    if object_action == "Update":  # check to see if provided information exists and is consistent with existing info
        term_guid = process_simple_attribute(txt, GUID_LABELS)
        update_description = process_simple_attribute(txt, ['Update Description'])
        term_display = (f"\n* Command: {command}\n\t* Glossary: {known_glossary_q_name}\n\t"
                        f"* Term Name: {term_name}\n\t* Qualified Name: {q_name}\n\t* Aliases: {aliases}\n\t"
                        f"* Categories: {categories}\n\t"
                        f"* Summary: {summary}\n\t* Description: {description}\n\t"
                        f"* Abbreviation: {abbreviation}\n\t* Examples: {examples}\n\t* Usage: {usage}\n\t"
                        f"* Version: {version}\n\t* Status: {status}\n\t* GUID: {term_guid}"
                        f"\n\t* Update Description: {update_description}\n")
        if not term_exists:
            msg = f"Update request invalid, Term {term_name} does not exist\n"
            print_msg(ERROR, msg, debug_level)
            valid = False

    elif object_action == 'Create':  # if the object_action is create, check that it doesn't already exist
        term_display = (f"\n* Command: {command}\n\t* Glossary: {known_glossary_q_name}\n\t"
                        f"* Term Name: {term_name}\n\t* Categories: {categories}\n\t* Summary: {summary}\n\t"
                        f"* Qualified Name: {q_name}\n\t* Aliases: {aliases}\n\t* Description: {description}\n\t"
                        f"* Abbreviation: {abbreviation}\n\t* Examples: {examples}\n\t* Usage: {usage}\n\t"
                        f"* Version: {version}\n\t* Status: {status}\n")
        if term_exists:
            msg = f"Term `{term_name}` cannot be created since it already exists\n"
            print_msg(ERROR, msg, debug_level)
        else:
            msg = f"It is valid to create Term `{term_name}`"
            print_msg(ALWAYS, msg, debug_level)

    if directive == "display":
        print(Markdown(term_display))
        return None
    elif directive == "validate":
        print(Markdown(term_display))
        return valid
    elif directive == "process":
        try:
            if not valid:  # First validate the term before we process it
                if term_exists and object_action == "Create":
                    msg = (f"Create failed because term `{term_name}` exists - changing `Create` to `Update` in "
                           f"processed output \n")
                    print_msg(ERROR, msg, debug_level)
                    print(Markdown(term_display))
                    return update_a_command(txt, object_action, object_type, known_q_name, known_guid)
                else:
                    return None

            print(Markdown(term_display))
            if object_action == "Update" and directive == "process":
                if not term_exists:
                    return None
                body = {
                    "class": "ReferenceableRequestBody", "elementProperties": {
                        "displayName": term_name, "class": "GlossaryTermProperties", "qualifiedName": known_q_name,
                        "aliases": alias_list, "summary": summary, "description": description,
                        "abbreviation": abbreviation, "examples": examples, "usage": usage,
                        "publishVersionIdentifier": version, "status": status
                        }, "updateDescription": update_description
                    }
                egeria_client.update_term(known_guid, body_slimmer(body), is_merge_update=False)
                # if cat_exist and cat_valid:
                update_term_categories(egeria_client, known_guid, cats_exist, cat_q_name_list)
                print_msg(ALWAYS,
                          f"\tUpdated Term `{term_name}` with GUID {known_guid}\n\tand categories `{categories}`",
                          debug_level)
                return egeria_client.get_term_by_guid(known_guid,
                                                      'md')  # return update_a_command(txt, object_action, object_type,
                # known_q_name, known_guid)
            elif object_action == "Update" and directive == "validate":  # is sthis reachable?
                return egeria_client.get_term_by_guid(known_guid, 'md')

            elif object_action == "Create":
                if term_exists:
                    msg = f"Term {term_name} exists so result document modifies term create to term update"
                    print_msg(INFO, msg, debug_level)
                    return update_a_command(txt, object_action, object_type, q_name, known_guid)
                else:
                    ## get the guid for the glossary from the name - first look in cache
                    cached = get_element_dictionary().get(known_glossary_q_name, None)

                    if cached is not None:
                        glossary_guid = cached.get('guid', None)
                        if glossary_guid is None:
                            msg = f"Glossary GUID for {known_glossary_q_name} not found in cache"
                            print_msg(WARNING, msg, debug_level)  # should this ever occur?
                            return None
                    else:
                        glossary_guid = egeria_client.__get_guid__(qualified_name=known_glossary_q_name)
                        if glossary_guid == NO_ELEMENTS_FOUND:
                            msg = f"Glossary {known_glossary_q_name} not found"
                            print_msg(WARNING, msg, debug_level)
                            return None
                    term_body = {
                        "class": "ReferenceableRequestBody", "elementProperties": {
                            "class": "GlossaryTermProperties", "qualifiedName": known_q_name, "displayName": term_name,
                            "aliases": alias_list, "summary": summary, "description": description,
                            "abbreviation": abbreviation, "examples": examples, "usage": usage,
                            "publishVersionIdentifier": version
                            # "additionalProperties":
                            #     {
                            #         "propertyName1": "xxxx",
                            #         "propertyName2": "xxxx"
                            #         }
                            }, "initialStatus": status
                        }
                    term_guid = egeria_client.create_controlled_glossary_term(glossary_guid, body_slimmer(term_body))
                    if term_guid == NO_ELEMENTS_FOUND:
                        print(f"{ERROR}Term {term_name} not created")
                        return None
                    if cats_exist and categories is not None:
                        update_term_categories(egeria_client, term_guid, cats_exist, cat_q_name_list)
                    update_element_dictionary(known_q_name, {'guid': term_guid, 'display_name': term_name})
                    print_msg(ALWAYS, f"Created term `{term_name}` with GUID {term_guid}", debug_level)
                    return egeria_client.get_term_by_guid(term_guid,
                                                          'MD')  # return update_a_command(txt, object_action,
                    # object_type, q_name, term_guid)
        except Exception as e:
            print(f"{ERROR}Error creating term {term_name}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_create_term_term_relationship_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> \
Optional[str]:
    """ Relate two terms through the specified relationship. ."""
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)
    object_type = command.split(' ')[1].strip()
    object_action = command.split(' ')[0].strip()
    term1_guid = None
    term2_guid = None

    term_relationship = process_simple_attribute(txt, ["Term Relationship", "Relationship Type"], "ERROR")
    if term_relationship not in TERM_RELATIONSHPS:
        valid = False

    print(Markdown(
        f"{pre_command} `{command}` for term relationship: `{term_relationship}` with directive: `{directive}` "))

    term1_q_name, term1_guid, term1_valid, term1_exists = process_element_identifiers(egeria_client, object_type,
                                                                                      ["Term 1 Name", "Term 1"], txt,
                                                                                      "Exists Required", None)

    term2_q_name, term2_guid, term2_valid, term2_exists = process_element_identifiers(egeria_client, object_type,
                                                                                      ["Term 2 Name", "Term 2"], txt,
                                                                                      "Exists Required", None)

    request_display = (f"\n\t* Term 1 Qualified Name: {term1_q_name}\n\t* Term 2 Qualified Name {term2_q_name}\n\t"
                       f"* Term Relationship: {term_relationship}")

    if not (term1_valid and term2_valid and term1_exists and term2_exists):
        valid = False

    if directive == "display":
        print(request_display)
        return None
    elif directive == "validate":
        print(request_display)
        return str(valid)
    elif directive == "process":
        try:
            print(request_display)
            if not valid:  # First validate the term before we process it
                return None
            egeria_client.add_relationship_between_terms(term1_guid, term2_guid, term_relationship)
            print_msg(ALWAYS, f"Relationship `{term_relationship}` created", debug_level)
            update_md = (f"\n\n# Update Term-Term Relationship\n\n## Term 1 Name:\n\n{term1_q_name}"
                         f"\n\n## Term 2 Name\n\n{term2_q_name}\n\n## Term Relationship:\n\n{term_relationship}")
            return update_md


        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_term_list_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)

    search_string = process_simple_attribute(txt, SEARCH_LABELS)
    if search_string is None:
        search_string = '*'
    print(Markdown(f"{pre_command} `{command}` with search string:`{search_string}` with directive: `{directive}`"))

    glossary = process_simple_attribute(txt, ['Glossary', 'In Glossary', "Glossary Name"])
    if glossary is not None:
        _, glossary_guid, _, glossary_exists = get_element_by_name(egeria_client, "Glossary", glossary)
        msg = f"Found glossary `{glossary}` with GUID {glossary_guid}"
        print_msg(INFO, msg, debug_level)
    else:
        glossary_guid = None
        msg = f"No glossary found"
        print_msg(INFO, msg, debug_level)

    output_format = process_simple_attribute(txt, OUTPUT_LABELS)
    if output_format is None:
        output_format = "LIST"
    elif output_format not in ELEMENT_OUTPUT_FORMATS:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = (f"\n\t* Search String: {search_string}\n\t* Glossary: {glossary}\n\t* Output Format: "
                       f"{output_format}\n")

    if directive == "display":
        print(Markdown(request_display))
        return None
    elif directive == "validate":
        print(Markdown(request_display))
        return valid
    elif directive == "process":
        try:
            print(Markdown(request_display))
            if not valid:  # First validate the term before we process it
                return None

            term_list_md = f"\n# Term List for search string: `{search_string}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.find_glossary_terms(search_string, glossary_guid, output_format=output_format)
                term_list_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                term_list_md += egeria_client.find_glossary_terms(search_string, glossary_guid,
                                                                  output_format=output_format)
            print_msg("ALWAYS", f"Wrote Term List for search string: `{search_string}`", debug_level)

            return term_list_md

            md_table = egeria_client.find_glossary_terms(search_string, glossary_guid, output_format=output_format)

            print_msg("ALWAYS", f"Wrote Term list for search string `{search_string}`", debug_level)
            return md_table

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_category_list_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)

    search_string = process_simple_attribute(txt, SEARCH_LABELS, "INFO")
    if search_string is None:
        search_string = '*'
    print(Markdown(f"{pre_command} `{command}` with search string:`{search_string}` with directive: `{directive}`"))

    output_format = process_simple_attribute(txt, OUTPUT_LABELS, "INFO")
    if output_format is None:
        output_format = "LIST"
    elif output_format not in ELEMENT_OUTPUT_FORMATS:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Search String: {search_string}\n\t* Output Format: {output_format}\n"

    if directive == "display":
        print(Markdown(request_display))
        return None
    elif directive == "validate":
        print(Markdown(request_display))
        return valid
    elif directive == "process":
        try:
            print(Markdown(request_display))
            if not valid:  # First validate the term before we process it
                return None

            cat_list_md = f"\n# Category List for search string: `{search_string}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.find_glossary_categories(search_string, output_format=output_format)
                cat_list_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                cat_list_md += egeria_client.find_glossary_categories(search_string, output_format=output_format)
            print_msg("ALWAYS", f"Wrote Category List for search string: `{search_string}`", debug_level)

            return cat_list_md

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:

        return None


def process_glossary_structure_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[
    str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)

    known_glossary_guid = ""

    glossary_name = process_simple_attribute(txt, GLOSSARY_NAME_LABELS, "ERROR")

    _, known_glossary_guid, valid, _ = process_element_identifiers(egeria_client, "Glossary", GLOSSARY_NAME_LABELS, txt,
                                                                   EXISTS_REQUIRED, None)

    print(Markdown(f"{pre_command} `{command}` for glossary:`{glossary_name}` with directive: `{directive}`"))

    output_format = process_simple_attribute(txt, OUTPUT_LABELS, "INFO")
    if output_format is None:
        output_format = "MD"
    elif output_format not in ["DICT", "LIST", "MD"]:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Glossary name: {glossary_name}\n\t* Output Format: {output_format}\n"

    if directive == "display":
        print(Markdown(request_display))
        return None
    elif directive == "validate":
        print(Markdown(request_display))
        return str(valid)
    elif directive == "process":
        try:
            print(Markdown(request_display))
            if not valid:  # First validate the term before we process it
                return None

            glossary_structure_md = f"\n# Glossary Structure for `{glossary_name}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.get_glossary_category_structure(known_glossary_guid, output_format=output_format)
                glossary_structure_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                glossary_structure_md += egeria_client.get_glossary_category_structure(known_glossary_guid,
                                                                                       output_format=output_format)
            print_msg("ALWAYS", f"Wrote Glossary Structure for glossary: `{glossary_name}`", debug_level)

            return glossary_structure_md

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_glossary_list_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)

    search_string = process_simple_attribute(txt, SEARCH_LABELS, "INFO")
    if search_string is None:
        search_string = '*'
    print(Markdown(f"{pre_command} `{command}` with search string:`{search_string}` with directive: `{directive}`"))
    if search_string is None:
        search_string = '*'

    output_format = process_simple_attribute(txt, OUTPUT_LABELS, "INFO")
    if output_format is None:
        output_format = "LIST"
    elif output_format not in ELEMENT_OUTPUT_FORMATS:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Search String: {search_string}\n\t* Output Format: {output_format}\n"

    if directive == "display":
        print(request_display)
        return None
    elif directive == "validate":
        print(request_display)
        return valid
    elif directive == "process":
        try:
            print(request_display)
            if not valid:  # First validate the term before we process it
                return None

            glossary_list_md = f"\n# Glossary List for `{search_string}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.find_glossaries(search_string, output_format=output_format)
                glossary_list_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                glossary_list_md += egeria_client.find_glossaries(search_string, output_format=output_format)
            print_msg("ALWAYS", f"Wrote Glossary List for search string: `{search_string}`", debug_level)

            return glossary_list_md

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_term_details_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)
    object_type = command.split(' ')[1].strip()
    object_action = command.split(' ')[0].strip()

    term_identifier = process_simple_attribute(txt, TERM_NAME_LABELS, "ERROR")

    print(Markdown(f"{pre_command} `{command}` for term:`{term_identifier}` with directive: `{directive}`"))

    output_format = process_simple_attribute(txt, OUTPUT_LABELS, "INFO")
    if output_format is None:
        output_format = "REPORT"
    else:
        output_format = output_format.upper()

    if output_format not in ["DICT", "REPORT"]:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Term Identifier: {term_identifier}\n\t* Output Format {output_format}"

    if directive == "display":
        print(request_display)
        return None
    elif directive == "validate":
        print(request_display)
        return valid
    elif directive == "process":
        try:
            print(request_display)
            if not valid:  # First validate the term before we process it
                return None
            output = egeria_client.get_term_details(term_identifier, output_format=output_format)
            if output_format == "DICT":
                output = f"```{json.dumps(output, indent=4)}```"
            print_msg("ALWAYS", f"Wrote Term Details for term: `{term_identifier}`", debug_level)

            return output

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_term_history_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[str]:
    """ List terms as a markdown table. Filter based on optional search string. """
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)
    object_type = command.split(' ')[1].strip()
    object_action = command.split(' ')[0].strip()

    element_labels = TERM_NAME_LABELS
    element_labels.append('Display Name')

    term_name = process_simple_attribute(txt, element_labels, "ERROR")

    known_q_name, known_guid, valid, term_exists = process_element_identifiers(egeria_client, object_type,
                                                                               element_labels, txt, object_action, )

    print(Markdown(f"{pre_command} `{command}` for term:`{term_name}` with directive: `{directive}`"))

    output_format = process_simple_attribute(txt, OUTPUT_LABELS, "INFO")
    if output_format is None:
        output_format = "LIST"
    elif output_format not in ["DICT", "LIST"]:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Term Name: {term_name}\n\t* Output Format {output_format}\n\t* GUID: {known_guid}\n"

    if directive == "display":
        print(request_display)
        return None
    elif directive == "validate":
        print(request_display)
        return valid
    elif directive == "process":
        try:
            print(request_display)
            if not valid:  # First validate the term before we process it
                return None
            term_history_md = f"\n# Term History for `{term_name}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.list_term_revision_history(known_guid, output_format=output_format)
                term_history_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                term_history_md += egeria_client.list_full_term_history(known_guid, output_format)
            print_msg("ALWAYS", f"Wrote Term History for term `{term_name}`", debug_level)

            return term_history_md

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None


def process_term_revision_history_command(egeria_client: EgeriaTech, txt: str, directive: str = "display") -> Optional[
    str]:
    """ List term revision history as a markdown table or list."""
    set_debug_level(directive)
    valid = True
    command = extract_command(txt)
    object_type = command.split(' ')[1].strip()
    object_action = command.split(' ')[0].strip()
    known_q_name = None
    known_guid = None

    element_labels = TERM_NAME_LABELS

    term_name = process_simple_attribute(txt, element_labels, "ERROR")
    print(Markdown(f"{pre_command} `{command}` for term: `{term_name}` with directive: `{directive}` "))

    known_q_name, known_guid, valid, _ = process_element_identifiers(egeria_client, object_type, element_labels, txt,
                                                                     object_action, )
    output_format = process_simple_attribute(txt, ['Output Format', 'Format'], 'INFO')
    if output_format is None:
        output_format = "LIST"
    elif output_format not in ["DICT", "LIST", "MD"]:
        valid = False
        print_msg(ERROR, f"Invalid output format: `{output_format}`", debug_level)

    request_display = f"\n\t* Term Name: {term_name}\n\t* Output Format: {output_format}\n"

    if directive == "display":
        print(request_display)
        return None
    elif directive == "validate":
        print(request_display)
        return str(valid)
    elif directive == "process":
        try:
            print(request_display)
            if not valid:  # First validate the term before we process it
                return None
            term_history_md = f"\n# Term Revision History for `{term_name}`\n\n"
            if output_format == "DICT":
                struct = egeria_client.list_term_revision_history(known_guid, output_format)
                term_history_md += f"```{json.dumps(struct, indent=4)}```\n"
            else:
                term_history_md += egeria_client.list_term_revision_history(known_guid, output_format)
            print_msg("ALWAYS", f"Wrote Term Revision History for term `{term_name}`", debug_level)
            return term_history_md

        except Exception as e:
            print(f"{ERROR}Error performing {command}: {e}")
            console.print_exception(show_locals=True)
            return None
    else:
        return None



"""PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module provides access to the metadata-explorer OMVS module.

[metadata-explorer](https://egeria-project.org/services/omvs/metadata-explorer/overview/)

"""

import asyncio

from httpx import Response

from pyegeria.utils import body_slimmer
from pyegeria._client import Client, max_paging_size
from pyegeria._globals import default_time_out, NO_ELEMENTS_FOUND



def query_seperator(current_string):
    if current_string == "":
        return "?"
    else:
        return "&"


"params are in the form of [(paramName, value), (param2Name, value)] if the value is not None, it will be added to the query string"


def query_string(params):
    result = ""
    for i in range(len(params)):
        if params[i][1] is not None:
            result = f"{result}{query_seperator(result)}{params[i][0]}={params[i][1]}"
    return result


def base_path(client, view_server: str):
    return f"{client.platform_url}/servers/{view_server}/api/open-metadata/metadata-explorer"


def process_related_element_list(
    response: Response, mermaid_only: bool, relationship_list: bool = False
) -> str | dict:
    """Process the result payload

    Parameters
    ----------
    response: Response
        - the response payload from the API call
    mermaid_only: bool
        - if true, only return the Mermaid graph
    relationship_list: bool
        - if True, look for "relationshipList" otherwise look for "relatedElementList"

    Returns
    -------

    """
    if relationship_list:
        elements = response.json().get("relationshipList", "No relationship list found")
    else:
        elements = response.json().get("relatedElementList", NO_ELEMENTS_FOUND)

    if isinstance(elements, str):
        return NO_ELEMENTS_FOUND
    if mermaid_only:
        return elements.get("mermaidGraph", "No mermaid graph found")

    el_list = elements.get("elementList", NO_ELEMENTS_FOUND)
    if isinstance(el_list, str):
        return el_list

    if len(el_list) == 0:
        return "No elements returned"
    return elements


class MetadataExplorer(Client):
    """MetadataExplorer is a class that extends the Client class. The Metadata Explorer OMVS provides APIs for
      supporting the search, query and retrieval of open metadata. It is an advanced API for users that understands
      the Open Metadata Types.

    Attributes:

        view_server_name: str
            The name of the View Server to connect to.
        platform_url : str
            URL of the server platform to connect to
        user_id : str
            The identity of the user calling the method - this sets a
            default optionally used by the methods when the user
            doesn't pass the user_id on a method call.
        user_pwd: str
            The password associated with the user_id. Defaults to None


    """

    def __init__(
        self,
        view_server: str,
        platform_url: str,
        user_id: str = None,
        user_pwd: str = None,
        token: str = None,
    ):
        self.view_server = view_server
        self.platform_url = platform_url
        self.user_id = user_id
        self.user_pwd = user_pwd
        self.metadata_explorer_command_root: str = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/metadata-explorer"
        )
        Client.__init__(
            self,
            view_server,
            platform_url,
            user_id=user_id,
            user_pwd=user_pwd,
            token=token,
        )

    #
    #   Get
    #

    async def _async_get_metadata_guid_by_unique_name(
        self,
        name: str,
        property_name: str,
        effective_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> str:
        """
        Retrieve the metadata element GUID using its unique name (typically the qualified name, but it is possible to
        specify a different property name in the request body as long as it is unique).
        If multiple matching instances are found, an exception is thrown. Async version.

        Parameters
        ----------
        name : str
            - unique name to search for
        property_name: str
            - property name to search in (typically the qualified name)
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        str
            The GUID of the element - or "No element found"

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "NameRequestBody",
            "name": name,
            "namePropertyName": property_name,
            "effectiveTime": effective_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/guid-by-unique-name"
            f"{possible_query_params}"
        )
        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body)
        )
        return response.json().get("guid", NO_ELEMENTS_FOUND)

    def get_metadata_guid_by_unique_name(
        self,
        name: str,
        property_name: str,
        effective_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> str:
        """
        Retrieve the metadata element GUID using its unique name (typically the qualified name, but it is possible to
        specify a different property name in the request body as long as it is unique).
        If multiple matching instances are found, an exception is thrown.

        Parameters
        ----------
        name : str
            - unique name to search for
        property_name: str
            - property name to search in (typically the qualified name)
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        str
            The GUID of the element - or "No element found"

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_metadata_guid_by_unique_name(
                name,
                property_name,
                effective_time,
                for_lineage,
                for_duplicate_processing,
            )
        )
        return response

    async def _async_get_metadata_element_by_guid(
        self,
        guid: str,
        effective_time: str = None,
        as_of_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> dict | str:
        """
        Retrieve the metadata element using its unique identifier. Async version.

        Parameters
        ----------
        guid : str
            - unique identifier of the element to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        as_of_time: str, default = None
            - Query the element as of this time. If None, then use current time.
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
            If the element is found, a dict of the element details is returned. Otherwise the string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "AnyTimeRequestBody",
            "effectiveTime": effective_time,
            "asOfTime": as_of_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/{guid}"
            f"{possible_query_params}"
        )
        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body)
        )
        return response.json().get("element", NO_ELEMENTS_FOUND)

    def get_metadata_element_by_guid(
        self,
        guid: str,
        effective_time: str = None,
        as_of_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> dict | str:
        """
        Retrieve the metadata element using its unique identifier.

        Parameters
        ----------
        guid : str
            - unique identifier of the element to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        as_of_time: str, default = None
            - Query the element as of this time. If None, then use current time.
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
            If the element is found, a dict of the element details is returned. Otherwise the string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_metadata_element_by_guid(
                guid, effective_time, for_lineage, for_duplicate_processing
            )
        )
        return response

    async def _async_get_anchored_metadata_element_graph(
        self,
        guid: str,
        effective_time: str = None,
        as_of_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        mermaid_only: bool = False,
    ) -> dict | str:
        """
        Retrieve the metadata element and all of its anchored elements using its unique identifier. Async version.

        Parameters
        ----------
        guid : str
            - unique identifier of the element to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        as_of_time: str, default = None
            - Query the element as of this time. If None, then use current time.
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        dict | str
            If the element is found, and mermaid_only is False, a dict of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "AnyTimeRequestBody",
            "effectiveTime": effective_time,
            "asOfTime": as_of_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/{guid}/with-anchored-elements"
            f"{possible_query_params}"
        )
        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body)
        )
        # if mermaid_only:
        #     return response.json()["elementGraph"].get(
        #         "mermaidGraph", NO_ELEMENTS_FOUND
        #     )
        # else:
        #     return response.json().get("elementGraph", NO_ELEMENTS_FOUND)
        if isinstance(response, str):
            return NO_ELEMENTS_FOUND
        el_list = response.json().get("elementGraph", NO_ELEMENTS_FOUND)
        if isinstance(el_list, str):
            return el_list
        if mermaid_only:
            return el_list.get("mermaidGraph", "No mermaid graph found")
        if len(el_list) == 0:
            return "No elements returned"
        return el_list

    def get_anchored_metadata_element_graph(
        self,
        guid: str,
        effective_time: str = None,
        as_of_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        mermaid_only: bool = False,
    ) -> dict | str:
        """
        Retrieve the metadata element and all of its anchored elements using its unique identifier. Async version.

        Parameters
        ----------
        guid : str
            - unique identifier of the element to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        as_of_time: str, default = None
            - Query the element as of this time. If None, then use current time.
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        dict | str
            If the element is found, and mermaid_only is False, a dict of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_anchored_metadata_element_graph(
                guid,
                effective_time,
                as_of_time,
                for_lineage,
                for_duplicate_processing,
                mermaid_only,
            )
        )
        return response

    async def _async_get_metadata_element_by_unique_name(
        self,
        name: str,
        property_name: str = "qualifiedName",
        effective_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> dict | str:
        """
        Retrieve the metadata element using its unique name (typically the *qualifiedName* attribute but other attributes
        can be used if they are unique - such as *pathName* for a file). Async version.

        Parameters
        ----------
        name : str
            - unique name to search for
        property_name: str, default = "qualifiedName"
            - property name to search in (typically the qualified name)
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
                If the element is found, a dict of the element details is returned. Otherwise the string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "NameRequestBody",
            "name": name,
            "namePropertyName": property_name,
            "effectiveTime": effective_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/by-unique-name"
            f"{possible_query_params}"
        )
        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body)
        )
        return response.json().get("element", NO_ELEMENTS_FOUND)

    def get_metadata_element_by_unique_name(
        self,
        name: str,
        property_name: str = "qualifiedName",
        effective_time: str = None,
        for_lineage: bool = False,
        for_duplicate_processing: bool = False,
    ) -> str:
        """
        Retrieve the metadata element using its unique name (typically the *qualifiedName* attribute but other attributes
        can be used if they are unique - such as *pathName* for a file).

        Parameters
        ----------
        name : str
            - unique name to search for
        property_name: str, default = "qualifiedName"
            - property name to search in (typically the qualified name)
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
                If the element is found, a dict of the element details is returned. Otherwise the string "No element found".


        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_metadata_element_by_unique_name(
                name,
                property_name,
                effective_time,
                for_lineage,
                for_duplicate_processing,
            )
        )
        return response

    async def _async_get_metadata_element_history(
        self,
        guid: str,
        effective_time: str = None,
        oldest_first: bool = False,
        from_time: str = None,
        to_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve all the versions of a metadata element. Async version.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        oldest_first: bool, default = False
        from_time: str, default = None
            Time to begin returning history
        to_time: str, default = None
            Time to end returning history
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("oldestFirst", oldest_first),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "HistoryRequestBody",
            "effectiveTime": effective_time,
            "fromTime": from_time,
            "toTime": to_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/{guid}/history"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        elements = response.json().get("elementList", NO_ELEMENTS_FOUND)
        if type(elements) is list:
            if len(elements) == 0:
                return NO_ELEMENTS_FOUND
        return elements

    def get_metadata_element_history(
        self,
        guid: str,
        effective_time: str = None,
        oldest_first: bool = False,
        from_time: str = None,
        to_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve all the versions of a metadata element.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        oldest_first: bool, default = False
        from_time: str, default = None
            Time to begin returning history
        to_time: str, default = None
            Time to end returning history
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_metadata_element_history(
                guid,
                effective_time,
                oldest_first,
                from_time,
                to_time,
                for_lineage,
                for_duplicate_processing,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_find_metadata_elements_with_string(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
    ) -> list | str:
        """
        Return a list of metadata elements that match the supplied criteria. The results can be returned over many pages.
        Async version.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            Returns a string if no elements found and a list of dict of elements with the results.

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "SearchStringRequestBody",
                  "searchString" : "Egeria",
                  "typeName" : "ValidValueDefinition",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }

        """

        possible_query_params = query_string(
            [
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/by-search-string"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        elements = response.json().get("elementList", NO_ELEMENTS_FOUND)
        if type(elements) is list:
            if len(elements) == 0:
                return NO_ELEMENTS_FOUND
        return elements

    def find_metadata_elements_with_string(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
    ) -> list | str:
        """
        Return a list of metadata elements that match the supplied criteria. The results can be returned over many pages.
        Async version.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            Returns a string if no elements found and a list of dict of elements with the results.

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "SearchStringRequestBody",
                  "searchString" : "Egeria",
                  "typeName" : "ValidValueDefinition",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_find_metadata_elements_with_string(
                body,
                for_lineage,
                for_duplicate_processing,
                start_from,
                page_size,
                time_out,
            )
        )
        return response

    async def _async_get_all_related_metadata_elements(
        self,
        guid: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the metadata elements connected to the supplied element.
        Async version.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned


        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".
        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """

        possible_query_params = query_string(
            [
                ("startingAtEnd", starting_at_end),
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/related-elements/{guid}/any-type"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )
        return process_related_element_list(response, mermaid_only)

    def get_all_related_metadata_elements(
        self,
        guid: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the metadata elements connected to the supplied element.

        Parameters
        ----------
        guid: str
            - unique identity of element
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned


        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_all_related_metadata_elements(
                guid,
                body,
                for_lineage,
                for_duplicate_processing,
                starting_at_end,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_get_related_metadata_elements(
        self,
        guid: str,
        relationship_type: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the metadata elements connected to the supplied element.
        Async version.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        relationship_type: str
            - name of relationship type to retrieve relationships of
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """

        possible_query_params = query_string(
            [
                ("startingAtEnd", starting_at_end),
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/related-elements/{guid}/type/{relationship_type}"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        return process_related_element_list(response, mermaid_only)

    def get_related_metadata_elements(
        self,
        guid: str,
        relationship_type: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the metadata elements connected to the supplied element.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        relationship_type: str
            - name of relationship type to retrieve relationships of
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_related_metadata_elements(
                guid,
                relationship_type,
                body,
                for_lineage,
                for_duplicate_processing,
                starting_at_end,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_get_all_metadata_element_relationships(
        self,
        end1_guid: str,
        end2_guid: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the relationships linking the supplied elements.
        Async version.

        Parameters
        ----------
        end1_guid: str
            - Unique identity of the metadata element at end1 of a relationship.
        end2_guid: str
            - Unique identity of the metadata element at end2 of a relationship.
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned


        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """

        possible_query_params = query_string(
            [
                ("startingAtEnd", starting_at_end),
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/{end1_guid}/linked-by-any-type/"
            f"to-elements/{end2_guid}"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        return process_related_element_list(response, mermaid_only)

    def get_all_metadata_element_relationships(
        self,
        end1_guid: str,
        end2_guid: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the relationships linking the supplied elements.

        Parameters
        ----------
        end1_guid: str
            - Unique identity of the metadata element at end1 of a relationship.
        end2_guid: str
            - Unique identity of the metadata element at end2 of a relationship.
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_all_metadata_element_relationships(
                end1_guid,
                end2_guid,
                body,
                for_lineage,
                for_duplicate_processing,
                starting_at_end,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_get_metadata_element_relationships(
        self,
        end1_guid: str,
        end2_guid: str,
        relationship_type: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the relationships linking the supplied elements.
        Async version.

        Parameters
        ----------
        end1_guid: str
            - Unique identity of the metadata element at end1 of a relationship.
        end2_guid: str
            - Unique identity of the metadata element at end2 of a relationship.
        relationship_type: str
            - name of relationship type to retrieve relationships of
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """

        possible_query_params = query_string(
            [
                ("startingAtEnd", starting_at_end),
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/{end1_guid}/linked-by-type/"
            f"{relationship_type}/to-elements/{end2_guid}"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        return process_related_element_list(response, mermaid_only)

    def get_metadata_element_relationships(
        self,
        end1_guid: str,
        end2_guid: str,
        relationship_type: str,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        starting_at_end: int = 0,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve the relationships linking the supplied elements.

        Parameters
        ----------
        end1_guid: str
            - Unique identity of the metadata element at end1 of a relationship.
        end2_guid: str
            - Unique identity of the metadata element at end2 of a relationship.
        relationship_type: str
            - name of relationship type to retrieve relationships of
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        starting_at_end: int, default = 0
            - Relationship end to start from.
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned


        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "ResultsRequestBody",
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "PROPERTY_ASCENDING",
                  "sequencingProperty": "fileName"
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_metadata_element_relationships(
                end1_guid,
                end2_guid,
                relationship_type,
                body,
                for_lineage,
                for_duplicate_processing,
                starting_at_end,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_find_metadata_elements(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
    ) -> list | str:
        """Return a list of metadata elements that match the supplied criteria.
        The results can be returned over many pages. Async version.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            Returns a string if no elements found and a list of dict of elements with the results.

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "FindRequestBody",
                  "metadataElementTypeName": "add typeName here",
                  "metadataElementSubtypeNames": [],
                  "searchProperties": {
                     "class" : "SearchProperties",
                     "conditions": [ {
                         "nestedConditions": {
                               "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "matchClassifications": {
                     "class" : "SearchClassifications",
                     "conditions": [{
                         "name" : "add classification name here",
                         "searchProperties": {
                            "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }
        """

        possible_query_params = query_string(
            [
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/metadata-elements/by-search-conditions"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        elements = response.json().get("elementList", NO_ELEMENTS_FOUND)
        if type(elements) is list:
            if len(elements) == 0:
                return NO_ELEMENTS_FOUND
        return elements

    def find_metadata_elements(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
    ) -> list | str:
        """
        Retrieve the relationships linking the supplied elements.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request

        Returns
        -------
        [dict] | str
            Returns a string if no elements found and a list of dict of elements with the results.

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "FindRequestBody",
                  "metadataElementTypeName": "add typeName here",
                  "metadataElementSubtypeNames": [],
                  "searchProperties": {
                     "class" : "SearchProperties",
                     "conditions": [ {
                         "nestedConditions": {
                               "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "matchClassifications": {
                     "class" : "SearchClassifications",
                     "conditions": [{
                         "name" : "add classification name here",
                         "searchProperties": {
                            "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_find_metadata_elements(
                body,
                for_lineage,
                for_duplicate_processing,
                start_from,
                page_size,
                time_out,
            )
        )
        return response

    async def _async_find_relationships_between_elements(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """Return a list of relationships that match the requested conditions.
            The results can be received as a series of pages. Async version.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "FindRelationshipRequestBody",
                  "relationshipTypeName": "add typeName here",
                  "searchProperties": {
                     "class" : "SearchProperties",
                     "conditions": [ {
                         "nestedConditions": {
                               "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }
        """

        possible_query_params = query_string(
            [
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        url = (
            f"{base_path(self, self.view_server)}/relationships/by-search-conditions"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )

        return process_related_element_list(
            response, mermaid_only, relationship_list=True
        )

    def find_relationships_between_elements(
        self,
        body: dict,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """Return a list of relationships that match the requested conditions.
            The results can be received as a series of pages.

        Parameters
        ----------
        body: dict
            - A structure containing the search criteria. (example below)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.

        Notes:

            Sample body:
                {
                  "class" : "FindRelationshipRequestBody",
                  "relationshipTypeName": "add typeName here",
                  "searchProperties": {
                     "class" : "SearchProperties",
                     "conditions": [ {
                         "nestedConditions": {
                               "class" : "SearchProperties",
                               "conditions": [
                               {
                                   "property" : "add name of property here",
                                   "operator": "EQ",
                                   "value": {
                                      "class" : "PrimitiveTypePropertyValue",
                                      "typeName" : "string",
                                      "primitiveValue" : "Add value here"
                                   }
                               }],
                               "matchCriteria": "ALL"
                         }
                     }],
                     "matchCriteria": "ANY"
                  },
                  "effectiveTime" : "{{$isoTimestamp}}",
                  "limitResultsByStatus" : ["ACTIVE"],
                  "asOfTime" : "{{$isoTimestamp}}",
                  "sequencingOrder": "CREATION_DATE_RECENT",
                  "sequencingProperty": ""
                }

        """
        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_find_relationships_between_elements(
                body,
                for_lineage,
                for_duplicate_processing,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response

    async def _async_get_relationship_by_guid(
        self,
        guid: str,
        effective_time: str = None,
        as_of_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> dict | str:
        """
        Retrieve the relationship using its unique identifier. Async version.

        Parameters
        ----------
        guid : str
            - unique identifier of the relationship to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
            If the relationship is found, a dict of the relationship details is returned. Otherwise, the string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "AnyTimeRequestBody",
            "effectiveTime": effective_time,
            "asOfTime": as_of_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/relationships/by-guid/{guid}"
            f"{possible_query_params}"
        )
        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body)
        )
        return response.json().get("element", NO_ELEMENTS_FOUND)

    def get_relationship_by_guid(
        self,
        guid: str,
        effective_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
    ) -> dict | str:
        """
        Retrieve the relationship using its unique identifier.

        Parameters
        ----------
        guid : str
            - unique identifier of the relationship to retrieve
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function

        Returns
        -------
        dict | str
            If the relationship is found, a dict of the relationship details is returned. Otherwise, the string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_relationship_by_guid(
                guid, effective_time, for_lineage, for_duplicate_processing
            )
        )
        return response

    async def _async_get_relationship_history(
        self,
        guid: str,
        effective_time: str = None,
        oldest_first: bool = False,
        from_time: str = None,
        to_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve all the versions of a relationship. Async version.

        Parameters
        ----------
        guid: str
            - Unique identity of element to retrieve.
        effective_time: str, default = None
            - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
        oldest_first: bool, default = False
        from_time: str, default = None
            Time to begin returning history
        to_time: str, default = None
            Time to end returning history
        for_lineage: bool, default is set by server
            - determines if elements classified as Memento should be returned - normally false
        for_duplicate_processing: bool, default is set by server
            - Normally false. Set true when the caller is part of a deduplication function
        start_from: int, default = 0
            - index of the list to start from (0 for start).
        page_size
            - maximum number of elements to return.
        time_out: int, default = default_time_out
            - http request timeout for this request
        mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

        Returns
        -------
        [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

        Raises
        ------
        InvalidParameterException
            one of the parameters is null or invalid or
        PropertyServerException
            There is a problem adding the element properties to the metadata repository or
        UserNotAuthorizedException
            the requesting user is not authorized to issue this request.
        """

        possible_query_params = query_string(
            [
                ("startFrom", start_from),
                ("pageSize", page_size),
                ("oldestFirst", oldest_first),
                ("forLineage", for_lineage),
                ("forDuplicateProcessing", for_duplicate_processing),
            ]
        )

        body = {
            "class": "HistoryRequestBody",
            "effectiveTime": effective_time,
            "fromTime": from_time,
            "toTime": to_time,
        }

        url = (
            f"{base_path(self, self.view_server)}/relationships/{guid}/history"
            f"{possible_query_params}"
        )

        response: Response = await self._async_make_request(
            "POST", url, body_slimmer(body), time_out=time_out
        )
        rel = response.json().get("relationshipList", NO_ELEMENTS_FOUND)
        if isinstance(rel, (list, dict)):
            return rel.get("elementList", NO_ELEMENTS_FOUND)
        else:
            return rel

    def get_relationship_history(
        self,
        guid: str,
        effective_time: str = None,
        oldest_first: bool = False,
        from_time: str = None,
        to_time: str = None,
        for_lineage: bool = None,
        for_duplicate_processing: bool = None,
        start_from: int = 0,
        page_size: int = max_paging_size,
        time_out: int = default_time_out,
        mermaid_only: bool = False,
    ) -> list | str:
        """
        Retrieve all the versions of a relationship.

         Parameters
         ----------
         guid: str
             - Unique identity of element to retrieve.
         effective_time: str, default = None
             - Time format is "YYYY-MM-DDTHH:MM:SS" (ISO 8601)
         oldest_first: bool, default = False
         from_time: str, default = None
             Time to begin returning history
         to_time: str, default = None
             Time to end returning history
         for_lineage: bool, default is set by server
             - determines if elements classified as Memento should be returned - normally false
         for_duplicate_processing: bool, default is set by server
             - Normally false. Set true when the caller is part of a deduplication function
         start_from: int, default = 0
             - index of the list to start from (0 for start).
         page_size
             - maximum number of elements to return.
         time_out: int, default = default_time_out
             - http request timeout for this request
         mermaid_only: bool, default is False
            - if true only a string representing the mermaid graph will be returned

         Returns
         -------
         [dict] | str
            If the element is found, and mermaid_only is False, a [dict] of the element details is returned.
            If mermaid_only is True, a string representing the mermaid graph will be returned.
            If no elements found, string "No element found".

         Raises
         ------
         InvalidParameterException
             one of the parameters is null or invalid or
         PropertyServerException
             There is a problem adding the element properties to the metadata repository or
         UserNotAuthorizedException
             the requesting user is not authorized to issue this request.
        """

        loop = asyncio.get_event_loop()
        response = loop.run_until_complete(
            self._async_get_relationship_history(
                guid,
                effective_time,
                oldest_first,
                from_time,
                to_time,
                for_lineage,
                for_duplicate_processing,
                start_from,
                page_size,
                time_out,
                mermaid_only,
            )
        )
        return response


if __name__ == "__main__":
    print("Main-Metadata Explorer")



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

General utility functions in support of the Egeria Python Client package.

"""

import json
from datetime import datetime

from rich import print, print_json
from rich.console import Console

console = Console(width=200)

# def wrap_text(df: pd.DataFrame, wrap_len: int = 30) -> pd.DataFrame:
#     """ Wrap the text in a dataframe
#     Parameters
#     ----------
#     df : pandas.DataFrame
#         The DataFrame to wrap the text in.
#     wrap_len : int, optional
#         The maximum length of each cell's contents to wrap. Defaults to 30.
#
#     Returns
#     -------
#     pandas.DataFrame
#         A new DataFrame with the wrapped text.
#
#     """
#
#     # Helper function to wrap text in a particular cell
#     def wrap_cell_contents(cell_contents: str, wrap: int = 30) -> str:
#         """ Wrap the text in a cell
#         Parameters
#         ----------
#         cell_contents : str or any
#             The contents of a cell in a dataframe.
#         wrap : int, optional
#             The maximum width at which to wrap the cell contents. Default is 30.
#
#         Returns
#         -------
#         str or any
#             If the cell_contents is a string and its length is greater than wrap_len,
#             the contents are wrapped at wrap_len width using textwrap.
#             Otherwise, the original cell_contents are returned as is.
#         """
#         if isinstance(cell_contents, str) and len(cell_contents) > wrap:
#             return textwrap.fill(cell_contents, width=wrap)
#         else:
#             return cell_contents
#
#     # Apply the helper function to each element in the DataFrame
#     return df.map(lambda x: wrap_cell_contents(x, wrap_len))
#
#
# def print_nice_table(df, wrap_len: int = 20, tablefmt: str = "grid") -> None:
#     """ Print a nice table from the data frame"""
#     print(tabulate(wrap_text(df, wrap_len), headers="keys", tablefmt=tablefmt))
#
#
# def get_json_as_table(input_json, wrap_len: int = 30, tablefmt: str = "grid") -> str:
#     """ return the input json as a table"""
#     data = json.loads(json.dumps(input_json))
#     df = pd.json_normalize(data)
#     return tabulate(wrap_text(df, wrap_len), headers="keys", tablefmt=tablefmt)
#
#
# def print_json_list_as_table(input_json, wrap_len: int = 30, tablefmt: str = "grid") -> None:
#     """ print a json list as a table"""
#     data = json.loads(json.dumps(input_json))
#     df = pd.json_normalize(data)
#     print(tabulate(wrap_text(df, wrap_len), headers="keys", tablefmt=tablefmt))


def print_rest_request_body(body):
    """

    Args:
        body:
    """
    pretty_body = json.dumps(body, indent=4)
    print_json(pretty_body, indent=4, sort_keys=True)


def print_rest_response(response):
    """

    Args:
        response:
    """
    print("Returns:")
    pretty_body = json.dumps(response, indent=4)
    print_json(pretty_body, indent=4, sort_keys=True)


def print_guid_list(guids):
    """Print a list of guids"""
    if guids is None:
        print("No assets created")
    else:
        pretty_guids = json.dumps(guids, indent=4)
        print_json(pretty_guids, indent=4, sort_keys=True)


#
# OCF Common services
# Working with assets - this set of functions displays assets returned from the open metadata repositories.
#


def print_response(response):
    """

    Args:
        response:

    Returns:
        : str
    """
    pretty_response = json.dumps(response.json(), indent=4)
    print(" ")
    print("Response: ")
    print(pretty_response)
    print(" ")


def print_unexpected_response(server_name, platform_name, platform_url, response):
    """

    Args:
        server_name:
        platform_name:
        platform_url:
        response:
    """
    if response.status_code == 200:
        related_http_code = response.json().get("related_http_code")
        if related_http_code == 200:
            print("Unexpected response from server " + server_name)
            print_response(response)
        else:
            exceptionErrorMessage = response.json().get("exceptionErrorMessage")
            exceptionSystemAction = response.json().get("exceptionSystemAction")
            exceptionUserAction = response.json().get("exceptionUserAction")
            if exceptionErrorMessage is not None:
                print(exceptionErrorMessage)
                print(" * " + exceptionSystemAction)
                print(" * " + exceptionUserAction)
            else:
                print("Unexpected response from server " + server_name)
                print_response(response)
    else:
        print(
            "Unexpected response from server platform "
            + platform_name
            + " at "
            + platform_url
        )
        print_response(response)


def get_last_guid(guids):
    """

    Args:
        guids:

    Returns:

    """
    if guids is None:
        return None
    else:
        return guids[-1]


def body_slimmer(body: dict) -> dict:
    """body_slimmer is a little function to remove unused keys from a dict

    Parameters
    ----------
    body : the dictionary that you want to slim

    Returns
    -------
    dict:
        a slimmed body
    """
    slimmed = {key: value for key, value in body.items() if value}
    return slimmed


if __name__ == "__main__":
    print("Main-Utils")
